''' SRCDS Entities Python module. '''
import warnings

import es
from . import srcds
from . import srcds_types

from .data import entities as entity_classes

ImplementationWarning = srcds.ImplementationWarning

# Constants:
# INVALID_INDEX is an index for entities that do not exist on the world
# initializing an entity with this value will cause the entity to be
# created and the new world index will be used for the entity
INVALID_INDEX = entity_classes.INVALID_INDEX
# NO_CREATE is a special value for entity initialization,
# the entity will not be created to the world and its index will stay
# as INVALID_INDEX; useful for entity templates
NO_CREATE = entity_classes.NO_CREATE # invalid index but do n


def get_entity(index):
    '''Return an entity class object for an existing entity.'''
    es.dbgmsg(2, 'Getting entity #%s' % (index,))
    classname = es.entitygetvalue(index, 'classname')
    es.dbgmsg(2, 'Classname: %r' % (classname,))
    if not classname:
        # entity does not exist
        raise srcds.SRCDSError('Entity #%s does not exist' % (index,))
    if classname in entity_class_lookup:
        t = entity_class_lookup[classname]
        ent = t(index)
        # ent._getvalues(index)
        return ent
    ent = entity_classes._UnknownEntity(index)
    # ent._getvalues(index)
    warnings.warn('unknown classname %r'%classname, ImplementationWarning, stacklevel=2) 
    return ent


def create_entity(classname):
    '''Create a new entity and return its instance.'''
    if classname not in entity_class_lookup:
        raise srcds.SRCDSError('Unknown entity class: %s' % (classname,))
    return entity_class_lookup[classname]()


class EntityFilter(object):
    '''Base class for entity filters.'''
    def __init__(self, name, value):
        self.name = name
        self.value = value

    __call__ = NotImplemented


class StringFilter(EntityFilter):
    '''Match with literal string.'''
    def __call__(self, index):
        real_value = es.entitygetvalue(index, self.name)
        return real_value == self.value


class WildcardFilter(EntityFilter):
    '''Match with wildcard string using * as the wildcard.'''
    def __init__(self, name, value, wildcard='*'):
        self.name = name
        self.parts = value.split(wildcard)
        self.start = len(self.parts[0])
        self.end = len(self.parts[-1])
    
    def __call__(self, index):
        real_value = es.entitygetvalue(index, self.name)
        es.dbgmsg(2, '(%s).%s = %r ? %r'%(index,self.name,real_value,self.parts))
        #real_value = index
        if real_value[:self.start] != self.parts[0]:
            es.dbgmsg(2, 'No start match')
            return False
        progress=self.start
        end = len(real_value)-self.end
        for part in self.parts[1:-1]:
            progress = real_value.find(part, progress, end)
            if progress == -1:
                es.dbgmsg(2, 'Part %s failed to match'%part)
                return False
        return real_value[len(real_value)-self.end:] == self.parts[-1]


class SpecialFilter(EntityFilter):
    '''Match with other types than strings.'''
    types = set(
        filter(
            lambda item: type(item) is type,
            vars(srcds_types).itervalues()
        )
    )

    def __init__(self, name, value):
        self.name = name
        self.value = value
        self.valuetype = type(value)
        if self.valuetype in self.types:
            # already a built-in type, can be converted to string
            strvalue = self.value.getvalue()
            strfilter = StringFilter(name, strvalue)
            self.__call__ = strfilter.__call__ # hax
        if issubclass(self.valuetype, EntityFilter):
            # it is a filter by itself, use it directly
            self.__call__ = self.value.__call__
        
    
    def __call__(self, index):
        # now to resolve the property, we need an instance
        try:
            entity = entity_class_lookup[classname](index)
        except KeyError:
            # we do not know of such entity, throw a warning
            entity = entity_classes._UnknownEntity(index)
        # get the type
        try:
            valuetype = entity._keyvalues[self.name]._type
            es.dbgmsg(2, 'type(%r.%r)=%r'%(entity, self.name, valuetype))
        except KeyError:
            # entity does not have such attribute, cannot match
            if type(entity) is entity_classes._UnknownEntity:
                warnings.warn(
                    'Cannot convert %r(%r) to built-in type for unimplemented '
                    'entity %r'%(self.value, self.valuetype, classname),
                    ImplementationWarning,
                    stacklevel=2
                )
                return False
            return False
        # attempt to convert
        strvalue = valuetype(self.value).getvalue()
        # speed up by reusing this
        strfilter = StringFilter(self.name, strvalue)
        self.__call__ = strfilter.__call__
        # match it!
        return strfilter(index)


def get_entities(classname=None, **kw):
    '''
    Return a dictionary containing entities matching the criteria.

    The dictionary has entity indices as keys and entity instances as values.
    
    Criteria can be given as keyword arguments and support multiple forms.
    A string is matched as is unless * is contained in it, in which case *
    acts as a wildcard. For some properties special values may be used,
    for example origin, angles and rendercolor accept 3-tuples which will
    be automatically converted to their proper string presentations.
    
    The first positional parameter corresponds to keyword classname. Usage of
    this parameter is encouraged when applicable. Limiting search by specific
    classname is more efficient than specifying wildcard classname or nothing
    at all.
    
    Some examples:
    all_entities = get_entities()
    players = get_entities('player')
    # alternatively get_entities(classname='player')
    all_props = get_entities('prop_*')
    # alternatively get_entities(classname='prop_*')
    red_physics_objects = get_entities('prop_physics', rendercolor=(255,0,0))
    '''
    filterchain = []
    if classname is None:
        indexlist = es.getEntityIndexes()
    elif '*' in classname:
        indexlist = es.getEntityIndexes()
        filterchain.append(WildcardFilter('classname', classname))
    else:
        indexlist = es.createentityindexlist(classname)
    for name, value in kw.iteritems():
        if '*' in value:
            filterchain.append(WildcardFilter(name, value))
        elif isinstance(value, basestring):
            filterchain.append(StringFilter(name, value))
        else:
            filterchain.append(SpecialFilter(name, value))
    return filter_entities(indexlist, filterchain)


def filter_entities(indexlist, filterchain):
    '''
    Return a dictionary containing entities pointed by indexlist,
    filtered by filterchain.
    '''
    ents = {}
    for index in indexlist:
        if all((
            filterproc(index)
            for filterproc in filterchain
        )):
            ents[index] = get_entity(index)
    return ents


def _create_lookup_table(module):
    '''
    Create a dictionary of form {'entity_classname': EntityClass, ...}
    from a module that contains entities.
    '''
    return dict(
        filter(
            lambda item: (
                item[0][0] != '_' and type(item[1]) is type and issubclass(
                    item[1],
                    entity_classes._BaseEntity
                )
            ),
            vars(module).iteritems()
        )
    )


def import_game(gamename, import_globals=None):
    '''
    Import game-specific entities into global lookup table.
    The entities are loaded from entities_gamename.py,
    if that does not exist, ImportError is raised.
    
    By default the file is looked from the srcds/data/ package, but this can
    be overridden by providing different import_globals object, for example
    srcds.entities.import_game('custom', globals()) will load
    entities_custom.py from the same package the module executing this
    command is from.
    
    Entity definitions that already exist are overwritten.

    Additionally, this function also returns the imported module (mostly for
    debugging purposes).
    '''
    gamemodule = __import__(
        'entities_'+gamename,
        globals=vars(entity_classes),
        level=1
    )
    game_entities = _create_lookup_table(gamemodule)
    entity_class_lookup.update(game_entities)
    # NOTE that this is not strictly allowed by Python and may break in future
    globals().update(game_entities)
    return gamemodule


# base entities
entity_class_lookup = _create_lookup_table(entity_classes)
# allow creating new entities by simply srcds.entities.entity_name()
# to do this, the entity class lookup dictionary is merged into global scope
# NOTE that this is not strictly allowed by Python and may break in future
vars().update(entity_class_lookup)

# Attempt to import game-specific entities for current game
try:
    import_game(es.getGameName())
except ImportError:
    warnings.warn('No game-specific entities implemented for this game.')

